 /***************************************************************
 *                                                              *
 * File: jxint.c                           Date: 12/13/94       *
 *                                                              *
 * Disclaimer:                                                  *
 *              Intel provides this AS IS, WITHOUT ANY WARRANTY,*
 *              INCLUDING THE WARRANTY OF MERCHANTABILITY OR    *
 *              FITNESS FOR A PARTICULAR PURPOSE, and makes no  *
 *              guarantee or representations regarding the use  *
 *              of, or the results of the use of, the software  *
 *              and documentation in terms of correctness,      *
 *              accuracy, reliability, currentness, or          *
 *              otherwise,  and you rely on the software,       *
 *              documentation, and results solely at your own   *
 *              risk.                                           *
 *                                                              *
 * Description:                                                 *
 *              The purpose of this program is to provide       *
 *              an example of the internal timers on the JX     *
 *              processor to generate interrupts.               *
 *                                                              *
 * Revision History:                                            *
 * ------------------------------------------------------------ *
 *                                                              *
 ****************************************************************/
#include <stdio.h>
#include "system.h"
#include "timers.h"

/* ensures that the processors state is retained and replaced    */
/* after the interrupt handler has completed.                    */
#pragma isr myvecint

/*            Function prototypes                                */
extern void myvecint(void);
void load_lock(void (*proc_addr)());

/* Define the user exit flag.  This flag is true if the user     */
/* opts to exit the program.                                     */
volatile static int exitflag;

/* Count the number of interrupts received from timer 0          */
int tmr0_count;


/*******************************************************************
 * File:  JXINT.C 	                    Date: 12/13/94         *
 *                                                                 *
 * Procedure: main                        Author: Intel Corp.      *
 *                                                                 *
 * Description:                                                    *
 *     This code will perform the following steps in handling of   *
 *     a internal JX timer interrupt.                              *
 *        1.  modify the ICON register                             *
 *        2.  modify the IMAP register                             *
 *        3.  caching of vector and interrupt handling procedure   *
 *        4.  lower processor priority                             *
 *        5.  modify the IMSK register                             *
 *        6.  clear the IPND register                              *
 *        7.  generate the timer interrupt                         *
 *                                                                 *
 *                                                                 *
 *******************************************************************/
int main ()
{                            /* begin main() */

                     /* Pointer to timer0 address */
TIMER_MMR_TYPE  *myptr0 = (TIMER_MMR_TYPE *)TMR0_MMR_BASE, 

                     /* Pointer to timer1 address */
		*myptr1 = (TIMER_MMR_TYPE *)TMR1_MMR_BASE;

unsigned long        /* Pointer to ICON address */
		*icon_addr  = (unsigned long *)ICON_MMR,

                    /* Pointer to IMAP2 address */
                *imap2_addr = (unsigned long *)IMAP2_MMR,

                    /* Timer variables */
		setup0, 
		setup1; 

                    /* Pointer to data ram memory */
unsigned int    *dram = (unsigned int *)0x0;

  
/*  1. Initialize global variables                                         */

       exitflag = 0;
       tmr0_count = 0;

/*  2. Set the ICON register:                                             */
/*     Dedicated interrupt, falling edge activated, global                */
/*     interrupts disabled, mov to r3 mask unchanged, fetch               */
/*     vector enable fast, fast sampling.                                 */

       *icon_addr = 0x67fc;
      
/*  3. Set the IMAP2 register:                                            */
/*     vector number E2 (226) for external interrupt xint6                */
       
       *imap2_addr = (*imap2_addr & 0x000000) | 0x0e0000;     

/*  4. Store the vector number in internal data ram memory                */ 
/*     Last 2 bits must end with 10 inorder to be cached properly.        */   

       dram[0xe] = (unsigned int)myvecint | (unsigned int)0x2; 

  
/*  5. Load and lock the interrupt procedure in the instruction           */ 
/*     cache.                                                             */
       
       system_control(CONFIGURE_ICACHE, LOCK_HALF_ICACHE, 0,
                                                    (void*)myvecint, 0);
        


/*  6. Setup timer0 to generate interrupts:  disable, auto reload,         */
/*     supervisor write, clock = bus/8                                     */

       setup0 = TMR_DISABLE | TMR_RLD_ENABLE | TMR_SUPCTL | TMR_TICK_PER_8BCLKS;
       timer_configure(myptr0,setup0,0x4c4b4,0x4c4b4);

/*  7. Setup timer1: disable, auto reload, all writes,                     */
/*     timer clock = bus clock                                             */

       setup1 = TMR_DISABLE | TMR_RLD_ENABLE | TMR_ALLCTL | TMR_TICK_PER_1BCLKS;
       timer_configure(myptr1,setup1,0x1312d00,0x1312d00);         


/*  8. Change processor priority to 0 so that our interrupt               */
/*     which is at priority 28 will be processed                          */

       modify_pc_priority(PRIORITY0);


/*  9. Set IMSK register to accept internal interrupts                     */
/*     timer0                                                              */

       i960_mod_reg(IMSK,0x1000);  

/* 10. Clear the interrupt pending register IPND                          */
    
       i960_mod_reg(IPND,0); 
        
/* 11. Enable global interrupts                                           */
        
       interrupt_enable();

        
/* 12. Enable both timers                                                 */

       timer_enable(myptr0,TMR_ENABLE);
       timer_enable(myptr1,TMR_ENABLE);

/*     Loop until exitflag is set                                         */
       while( !exitflag )  
	;             /* empty statement list */

}   /* end main() */



/****************************************************************
 *                                                              *
 * Procedure: myvecint                  Date: 12/13/94          *
 *                                                              *
 * File: jxint.c                        Author: Intel Corp.     *
 *                                                              *
 * Description:                                                 *
 *              This routine is the interrupt service procedure *
 *              to be called by the processor everytime the     *
 *              timer0 count expires.  It checks the count      *
 *              of timer1 and if timer1 has counted down to     *
 *              to 0x300000 or lower,  then prompts the user    *
 *              if he/she would like to exit the program.       *
 *                                                              *
 *     Inputs - None                                            *
 *    Outputs - None                                            *
 *                                                              *
 ****************************************************************/
void myvecint()
{                /* begin myvecint() */

                /* Temporary variables */
int             count ;   
char            c;

                /* pointer to timer0 */
TIMER_MMR_TYPE  *myptr0 = (TIMER_MMR_TYPE *)TMR0_MMR_BASE, 
                /* pointer to timer1 */
		*myptr1 = (TIMER_MMR_TYPE *)TMR1_MMR_BASE; 


    /*  Disable both timers  */
    timer_enable(myptr0,TMR_DISABLE);
    timer_enable(myptr1,TMR_DISABLE);

    /* Checks count and if want to continue */
    count = myptr1->TCR;

    /* display count */
    printf("    Interrupt %d info:\n", tmr0_count++);
    printf("        TMR0 count: 0x%x\n", myptr0->TCR); 
    printf("        TMR1 count: 0x%x\n", count);

    /* check count */
    if( count <= 0x300000 )
      { 
        printf("PAUSE, hit <Enter> key to continue or any other key to Exit: ");
           
        /* check if want to continue */
        if(( c=getchar() ) == '\n' )
          {  
            timer_enable(myptr0,TMR_ENABLE);
            timer_enable(myptr1,TMR_ENABLE);
          }    
            
        /* set exit flag and return to main */
        else
          {  
            while(( c = getchar() ) != '\n')   
              ;	 
            exitflag = 1;
          }  
      }   
       
    /* count not ready */
    else
      {  
        timer_enable(myptr0,TMR_ENABLE);
        timer_enable(myptr1,TMR_ENABLE);
      }   
  
}   /* end myvecint() */


 
