/****************************************************************
 *                                                              *
 * File: timers.c                       Date: 12/13/94          *
 *                                                              *
 * Disclaimer:                                                  *
 *              Intel provides this AS IS, WITHOUT ANY WARRANTY,*
 *              INCLUDING THE WARRANTY OF MERCHANTABILITY OR    *
 *              FITNESS FOR A PARTICULAR PURPOSE, and makes no  *
 *              guarantee or representations regarding the use  *
 *              of, or the results of the use of, the software  *
 *              and documentation in terms of correctness,      *
 *              accuracy, reliability, currentness, or          *
 *              otherwise,  and you rely on the software,       *
 *              documentation, and results solely at your own   *
 *              risk.                                           *
 *                                                              *
 * Description:                                                 *
 *              The purpose of this file is to provide the      *
 *              functions necessary to manipulate the JX        *
 *              Microprocessor timers.                          *
 *                                                              *
 * Revision History:                                            *
 * ------------------------------------------------------------ *
 *                                                              *
 ****************************************************************/
#include <stdio.h>
#include "timers.h"

 


/****************************************************************
 *                                                              *
 * Procedure: timer_configure           Date: 12/13/94          *
 *                                                              *
 * File: timers.c                        Author: Intel Corp.    *
 *                                                              *
 * Description:                                                 *
 *              This routine configures a timer by writing the  *
 *              control word, count and reload value if         *
 *              applicable to the timer MMR registers.          *
 *                                                              *
 *     Inputs - tptr    points to timer0 or timer1 address      *
 *              ctlwd   specifies if timer0 or timer1           *
 *              count   time left on timer1                     *
 *              reldcnt time to be reloaded                     *
 *    Outputs - None                                            *
 *                                                              *
 ****************************************************************/
void timer_configure(TIMER_MMR_TYPE *tptr, long ctlwd,
		     long count, long reldcnt)
{   /* begin timer_configure() */


          
    tptr->TMR = 0;                        /* disable timer first */

               
    tptr->TCR = count;
    if (ctlwd & 0x4) 
	tptr->TRR = reldcnt;         /* If Reload enabled, load */
    tptr->TMR = ctlwd;

}   /* end timer_configure() */



/****************************************************************
 *                                                              *
 * Procedure: timer_enable              Date: 12/13/94          *
 *                                                              *
 * File: timers.c                        Author: Intel Corp.    *
 *                                                              *
 * Description:                                                 *
 *              This routine enables the timer by setting the   *
 *              timer enable bit in the timer register.         *
 *                                                              *
 *     Inputs - tptr    pointer to timer0 or timer1             *   
 *              enable  value to set bit in TMR                 *
 *    Outputs - None                                            *
 *                                                              *
 ****************************************************************/
void timer_enable(TIMER_MMR_TYPE *tptr, long enable)
{              /* begin timer_enable() */
long temp;

    
    temp = tptr->TMR;           /* save old value in temp */
    tptr->TMR = (temp & ~0x2) | (enable & 0x2);  /* load new value */
     
}   /* end timer_enable() */



/****************************************************************
 *                                                              *
 * Procedure: timer_expired             Date: 12/13/94          *
 *                                                              *
 * File: timers.c                        Author: Intel Corp.    *
 *                                                              *
 * Description:                                                 *
 *              This function returns whether the timer has     *
 *              expired (1) or not (0) by checking the terminal *
 *              count bit in the timer mode register.           *
 *                                                              *
 *     Inputs - tptr pointer to timer0 or timer1                *
 *    Outputs - 1 timer has expired                             *
 *              0 timer has not expired                         *
 *                                                              *
 ****************************************************************/
int timer_expired(TIMER_MMR_TYPE *tptr)
{   /* begin timer_expired() */

    return (tptr->TMR & 0x1);

}   /* end timer_expired() */



/****************************************************************
 *                                                              *
 * Procedure: timer_start               Date: 12/13/94          *
 *                                                              *
 * File: timers.c                        Author: Intel Corp.    *
 *                                                              *
 * Description:                                                 *
 *              This routine restarts the timer by disabling    *
 *              the timer first,  reloading the count register  *
 *              and then enabling the timer.                    *
 *                                                              *
 *     Inputs - tptr    pointer to timer0 or timer1             *
 *              count   time left on timer0 or timer1           *
 *    Outputs - None                                            *
 *                                                              *
 ****************************************************************/
void timer_start(TIMER_MMR_TYPE *tptr, long count)
{                /* begin timer_start() */
long temp;

    temp = tptr->TMR;             /* store old TMR value */
    tptr->TMR = temp & ~0x3;      /* load new TMR value  */
    tptr->TCR = count;            /* load count into TCR */
    tptr->TMR = tptr->TMR | 0x2;

}   /* end timer_start() */
