/**************************************************
 *                                                *
 *  Module Name:  init.c                          *
 *                                                *
 *  Functions in this Module:                     *
 *    main()                                      *
 *                                                *
 *************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "arch960.h"
#include "ostypes.h"
#include "globals.h"
#include "utils.h"
#include "timer.h"
#include "init.h"
#include "inline.h"
#include "tomos.h"
#include "tasks.h"

/**************************************************
 *                                                *
 * Function:  main()                              *
 *                                                *
 * Description:                                   *
 *   main() is the first function called, and is  *
 *   the starting point of TOMOS.  main() init-   *
 *   ializes the interrupt controller so that     *
 *   a timer (tick) interrupt will be handled     *
 *   correctly.  The tick is the interrupt which  *
 *   will cause the context switch.  Main also    *
 *   installs the timer tick ISR, and lowers the  *
 *   processor priority so that the interrupts    *
 *   will be recognized.                         *
 *                                                *
 * Prototype:  int main(int argc, char *argv[])   *
 *                                                *
 * Parameters:                                    *
 *   argc        Stadard C argument containing    *
 *               the number of arguments on the   *
 *               command line.                    *
 *   argv        Standard C argument containing   *
 *               an array of pointers to strings  *
 *               which contain each command line  *
 *               argument.                        *
 *                                                *
 *   Currently, there are no command line         *
 *   arguments.                                   *
 *                                                *
 * Returns:                                       *
 *   Nothing.  This function never returns.       *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  12/30/94  Creation                       *
 *   TJ  02/06/95  Updated and commented.         *
 *                                                *
 *************************************************/
void main(int argc, char *argv[])
{
  PRCB *pPRCB;
  WORD wTmp, ii, IPND, IMSK;

  /*******************************************************
   * For the definition of the C images of the arch-     *
   * itecturally define initialization data structures,  *
   * please refer to ARCH960.H.  This header file        *
   * describes PRCB, CtlTblBase, etc.                    *
   ******************************************************/
  
  /*******************************************************
   * Get the current Processor Control Block (PRCB)      *
   * pointer.  This is a wrapper function for a function *
   * provided by MON960 for the Cx.  The MON960 function *
   * is the only way to get the PRCB pointer on the Cx   *
   * because only MON960 knows where it put the PRCB.    *
   *                                                     *
   * On the Jx and Hx, this function takes advantage of  *
   * the memory-mapped control register which stores the *
   * current PRCB pointer, and is therefore not depend-  *
   * ent on MON960.  Assuming the user defines the       *
   * architecture correctly on the compiler command line *
   * using the -Axx Switch, this code will compile       *
   * correctly for the specified target CPU.             *
   ******************************************************/
  pPRCB = GetPRCB();

  /* globally disable interrupts */
  DisableInts();

  /*******************************************************
   * SetProcPriority() executes the modpc instruction    *
   * in order to lower the processor priority.  The      *
   * timer tick interrupt will be mapped to vector 226,  *
   * which is at priority 28 (226 % 8).  After reset,    *
   * the processor is at priority 31 (assuming MON960    *
   * doesn't change that, but why assume?).  In order    *
   * for the timer tick interrupt to be recognized, the  *
   * processor priority must be lowered - to at least 27 *
   * in this case.                                       *
   ******************************************************/
  SetProcPriority(27);
 
  /*******************************************************
   * On the Cyclone Evaluation board, the CIO (timer)    *
   * output is connected to XINT6#.  We must now update  *
   * the IMAP1 register, to map the XINT6# pin to        *
   * interrupt vector 226.  This is accomplished by      *
   * writing 0xE to the bit field in IMAP1 corresponding *
   * to XINT6#.  This effectively maps XINT6# to vector  *
   * 0xE2 (or 226).                                      *
   *                                                     *
   * On the Cx we don't actually write to the register,  *
   * but instead update an image of the register in      *
   * memory, and then execute a sysctl instruction to    *
   * read the memory image and actually write the        *
   * regsiter.  The Jx and Hx provide memory-mapped      *
   * control registers; this means that IMAPx can be     *
   * more simply accessed (with restrictions) using      *
   * normal ld/st instructions.                          *
   ******************************************************/
  
  /* update Interrupt Mapping register IMAP1 */
  pPRCB->CtlTblBase->IMAP1.Bits.X2 = IMAP_TO_VEC226;

  /*******************************************************
   * We must also write a pointer to the interrupt       *
   * service routine into the interrupt table in the     *
   * entry for vector 226.  The processor fetches this   *
   * pointer when the interrupt controller determines    *
   * that the interrupt vector 226 must be serviced.     *
   ******************************************************/
  
  /* install ISR for vector 226 */
  pPRCB->IntTblBase->IntVecs0[ENTRY_VEC226].ISR = Tick;

  /*******************************************************
   * ReadIPND(), ReadIMSK(), WriteIPND(), and            *
   * WriteIMSK() are functions that access the Interrupt *
   * Pending (IPND) and Interrupt Mask (IMSK) registers  *
   * by accessing these registers as special function    *
   * registers (SFRs) on the Cx, or using the memory-    *
   * mapped control register capability on the Jx and    *
   * Hx.  Note that the Jx does not contain special      *
   * function registers.  See UTILS.C for the definition *
   * of these functions.                                 *
   ******************************************************/
  
  /* read IMSK and IPND registers */
  IPND = ReadIPND();
  IMSK = ReadIMSK();

  /*******************************************************
   * We just read the IPND and IMSK registers.  We will  *
   * be using dedicated mode interrupts, meaning the     *
   * XINTx# pins will act as direct interrupt inputs,    *
   * and will generate an interrupt based on the         *
   * programming of the IMAPx registers.  We must now    *
   * unmask the XINT6# interrupt so the timer tick can   *
   * `get in.'                                           *
   ******************************************************/
  WriteIMSK(IMSK | 0x40);

  /*******************************************************
   * The ability to write the IPND register was used     *
   * during debug.  This is a useful means of generating *
   * one (and only one) interrupt through software.      *
   * This code is commented out and should remain so     *
   * except for debug.                                   *
   ******************************************************/
  /* WriteIPND(IPND | 0x40); */

  /*******************************************************
   * Configure the timer period with InitTimerTick().    *
   * The function's argument sets the clock period of    *
   * the timer.                                          *
   ******************************************************/
  InitTimerTick(0, 0x100);

  /* initialize TOMOS task switcher - also creates background task */
  TOMOSInit();

  /* create a user task - something to run besides the background task */
  TaskCreate(MyTask, 0, TASK_STK_SZ, TASK_PRIOR, TRUE);
  
  /* start TOMOS and begin multitasking */
  StartTOMOS();
}

