/**************************************************
 *                                                *
 *  Header:  inline.h                             *
 *                                                *
 *  A collection of inline functions.             *
 *                                                *
 *************************************************/

#ifndef _INLINE_H_
#define _INLINE_H_

#include <stdlib.h>
#include "types.h"
#include "arch960.h"
#include "tomos.h"
#include "globals.h"
#include "utils.h"

#pragma inline DisableInts EnableInts StartHighestPriorRdy

/**************************************************
 *                                                *
 * Function:  DisableInts()                       *
 *                                                *
 * Description:                                   *
 *   Globally disables interrupts by setting the  *
 *   global interrupt enable bit (gie) of the     *
 *   interrupt control register (ICON).           *
 *                                                *
 * Prototype:  void DisableInts(void)             *
 *                                                *
 * Parameters:                                    *
 *   None.                                        *
 *                                                *
 * Returns:                                       *
 *   Nothing.                                     *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  12/30/94  Creation                       *
 *                                                *
 *************************************************/
inline static void DisableInts()
{
  #if defined(__i960CA__) || defined(__i960CF__)
  register PRCB *pPRCB;
  register int iTmp asm("r3");

  /* get the old value of ICON from control table */
  pPRCB = (PRCB *) GetPRCB();
  
  /* if interrupts are already disabled, return */
  if (pPRCB->CtlTblBase->ICON.Bits.GlbIntEn)
    return;

  /* set the global enable bit - ie, disable interrupts */
  pPRCB->CtlTblBase->ICON.Bits.GlbIntEn = 1;

  /* execute the sysctl instruction make new ICON effective */
  iTmp = 0x401;
  asm volatile ("sysctl %0, 0, %1" : : "l" (iTmp), "l" (iTmp));

  #elif defined(__i960JA__) || \
        defined(__i960JD__) || \
        defined(__i960JF__) || \
        defined(__i960HA__) || \
        defined(__i960HD__) || \
        defined(__i960HT__)

  asm volatile ("intdis");

  #else
  #error Unsupported Architecture
  #endif

  /* consume cycles and let changes take effect */
  asm volatile ("mov r4, r4");
  asm volatile ("mov r4, r4");
  asm volatile ("mov r4, r4");
}

/**************************************************
 *                                                *
 * Function:  EnableInts()                        *
 *                                                *
 * Description:                                   *
 *   Globally enables interrupts by clearing the  *
 *   global interrupt enable bit (gie) of the     *
 *   interrupt control register (ICON).           *
 *                                                *
 * Prototype:  void EnableInts(void)              *
 *                                                *
 * Parameters:                                    *
 *   None.                                        *
 *                                                *
 * Returns:                                       *
 *   Nothing.                                     *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  12/30/94  Creation                       *
 *                                                *
 *************************************************/
inline static void EnableInts()
{
  #if defined(__i960CA__) || defined(__i960CF__)
  register PRCB *pPRCB;
  register int iTmp asm("r3");

  /* get the old value of ICON from control table */
  pPRCB = (PRCB *) GetPRCB();
  
  /* if interrupts already enabled return */
  if (!(pPRCB->CtlTblBase->ICON.Bits.GlbIntEn))
    return;

  /* clear the global enable bit - ie, enable interrupts */
  pPRCB->CtlTblBase->ICON.Bits.GlbIntEn = 0;

  /* execute the sysctl instruction make new ICON effective */
  iTmp = 0x401;
  asm volatile ("sysctl %0, 0, %1" : : "l" (iTmp), "l" (iTmp));

  #elif defined(__i960JA__) || \
        defined(__i960JD__) || \
        defined(__i960JF__) || \
        defined(__i960HA__) || \
        defined(__i960HD__) || \
        defined(__i960HT__)


  asm volatile ("inten");

  #else
  #error Unsupported Architecture
  #endif

  /* consume three cycles and let changes take effect */
  asm volatile ("mov r4, r4");
  asm volatile ("mov r4, r4");
  asm volatile ("mov r4, r4");
}

/**************************************************
 *                                                *
 * Function:  StartHighestPriorRdy()              *
 *                                                *
 * Description:                                   *
 *   Starts the highest priority task ready to    *
 *   execute by preparing the stack frame to      *
 *   which the task will return.  Processor is    *
 *   made to return to this particular stack      *
 *   frame by writing the pfp register.           *
 *                                                *
 * Prototype:  void StartHighestPriorRdy(void)    *
 *                                                *
 * Parameters:                                    *
 *   None.                                        *
 *                                                *
 * Returns:                                       *
 *   Nothing.                                     *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  12/30/94  Creation                       *
 *                                                *
 *************************************************/
inline static void StartHighestPriorRdy()
{
  OSTCB *pTCB;
  register WORD wTmp;
  register WORD *pwGlblPtr asm("r4");

  /* get pointer to TCB of highest priority task ready to run */
  pTCB = aptcbActiveTasks[wHighestPriorRdy];
  
  /* set current task pointer to point to this TCB */
  ptcbCur = pTCB;

  /* change task state to running */
  pTCB->tsStatus = TASK_RUNNING;

  /* restore control (AC, TC, PC) registers */
  WriteAC(pTCB->ArithCtl);
  WritePC(pTCB->ProcCtl);
  WriteTC(pTCB->TraceCtl);

  /* update pfp and return-type field                         */
  /* DON'T do interrupt return - this is just a normal return */

  asm volatile ("flushreg");
  
  wTmp = pTCB->wPrevFrmPtr;
  wTmp &= ~0xF;         /* clear return type bits for local return */
  asm volatile ("mov %0, pfp" : : "dI" (wTmp));

  asm volatile ("flushreg");

  /*******************************************************
   * Notice the second flushreg instruction.  With the   *
   * pipelined architecture of the 80960, it is possible *
   * to modify the pfp register, immediately execute a   *
   * return instruction, and have the processor pick up  *
   * the old (unmodified) value of the pfp upon return.  *
   * Executing the second flushreg forces the processor  *
   * to wait until all modifications to the pfp are      *
   * complete before continuing execution.  Since the    *
   * register cache is already flushed, this second      *
   * flushreg will execute -VERY- fast.                  *
   *                                                     *
   * While in this case, the second flushreg is not      *
   * manditory, it is included here for documentation    *
   * purposes.                                           *
   ******************************************************/

  /* restore global registers */
  pwGlblPtr = &pTCB->GlobalRegs.g0;

  asm volatile ("ldq (%0), g0"    : : "l" (pwGlblPtr));
  asm volatile ("addo 16, %0, %0" : : "l" (pwGlblPtr));

  asm volatile ("ldq (%0), g4"    : : "l" (pwGlblPtr));
  asm volatile ("addo 16, %0, %0" : : "l" (pwGlblPtr));

  asm volatile ("ldq (%0), g8"    : : "l" (pwGlblPtr));
  asm volatile ("addo 16, %0, %0" : : "l" (pwGlblPtr));

  asm volatile ("ldq (%0), g12"   : : "l" (pwGlblPtr));

  /*******************************************************
   * When TOMOS starts, the implicit return instruction  *
   * at the end of this function causes a return to      *
   * IdleTask().  In IdleTask(), interrupts are enabled  *
   * and multitasking begins.                            *
   ******************************************************/
}

#endif
