/**************************************************
 *                                                *
 *  Module Name:  tick.s                          *
 *                                                *
 *  The assembly routine is executed with each    *
 *  timer tick.  This routine manages delayed     *
 *  tasks, determines when a context switch is    *
 *  necessary, and performs switch when required. *
 *                                                *
 *************************************************/

/* constant definitions */
.equ TASK_READY,        0
.equ TASK_RUNNING,      1

/* make Tick identifier visible to linker */
.globl _Tick

/* force alignement to quad-word boundary */
.align 4

_Tick:

        /****************************************
         * Increment the timer tick counter     *
         * (wTicker).                           *
         ***************************************/
        ld      _wTicker, r3
        addo    1, r3, r4
        st      r4, _wTicker

        /****************************************
         * Each timer tick generated by the CIO *
         * must be acknowledged.  This is more  *
         * magic code abtained from Walt Gorman *
         * in Oregon to acknowledge timer 0     *
         * events.  In the future, this code    *
         * will be documented more completely.  *
         ***************************************/
        ld      _CIO, r3
        ldconst 0x0A, r4
        ldconst 0x24, r6

        stob    r4, 12(r3)
        ldob    12(r3), r5
        stob    r4, 12(r3)
        stob    r6, 12(r3)

        /****************************************
         * In the task control block (TCB) each *
         * task knows how much stack space it   *
         * was allocated.  Do a little checking *
         * here to make sure a task has not     *
         * exceeded its stack space.  Fairly    *
         * simple yet useful operation...worth  *
         * the overhead.  Depending on how bad  *
         * things are screwed up, the error     *
         * message may or may not print out in  *
         * a readable form.  Nevertheless, the  *
         * system will shut down in a sensible  *
         * fashion.                             *
         ***************************************/
        ld      _ptcbCur, r3  /* get pointer to current task       */
        ldl     0x54(r3), r4  /* r4 = pvAllocStack r5 = wStackSize */
        addo    r4, r5, r6    /* stack shouldn't exceed r4 + r5    */
        cmpobl.t pfp, r6, f1  /* stack exceeded?                   */
        ldconst _stack_exceeded, g0
                              /* stack exceeded - print error msg  */
        ldob    _wHighestPriorRdy, g1
                              /* and priority of offending task    */
        callj   _printf       /* call printf() to display error    */
        ldconst -1, g0        /* return -1 -warning danger Will R. */
        callj   _exit         /* call exit()                       */
        

f1:
        /****************************************
         * Now we must check for delayed tasks  *
         * and decrement their delay counters.  *
         * If any time out (reach zero) we must *
         * clear the corresponding bit in the   *
         * delayed task field, and set the cor- *
         * responding bit in the task priority  *
         * field.                               *
         ***************************************/
        
        /****************************************
         * Get base addresses of array of       *
         * pointers to TCBs, and delayed task   *
         * field                                *
         ***************************************/
        ldconst _aptcbActiveTasks, r3
        ldconst _abTaskDelayed, r4

        /****************************************
         * Initialize constants so we can       *
         * iterate through the bit field.       *
         ***************************************/
        ldconst 0, r5   /* bit to check                 */
        ldconst 0, r6   /* word being checked           */
        ldconst 32, r7  /* multiplier used for indexing */

        /* load the first quad-word of abTaskDelayed */
        ldq     (r4), r12
        
        /* check for set bits */
check_a_bit:
        bbs.f   r5, r12, delayed_task_found
                                /* check for delayed task       */
bump_bit_count:
        addo    1, r5, r5       /* increment position counter   */
        cmpobl.t r5, r7, check_a_bit
                                /* if less than 32 check more   */
        ldconst 0, r5           /* zero bit position counter    */
        addo    1, r6, r6       /* increment word counter       */
        cmpobe.f 8, r6, delays_done
                                /* done finding delayed tasks   */
        cmpobne.t 4, r6, get_next_word
                                /* need to load another quad?   */
        ldq     0x10(r4), r12   /* get next quad-word of delays */
        b       check_a_bit     /* continue                     */        

        /****************************************
         * Data from the abTaskDelayed array is *
         * loaded a quad-word at a time.  But   *
         * to make the loop simpler, we always  *
         * want to check for set bits in one    *
         * register (r12 in this case).  This   *
         * next bit of code moves the next word *
         * of the quad-word into the proper     *
         * register so scanning can continue.   *
         ***************************************/
get_next_word:
        ld      register_move_table[r6*4], r9
                                /* get branch address           */
        bx      (r9)            /* branch to correct operation  */

register_move_table:
        .word   0
        .word   r13_to_r12
        .word   r14_to_r12
        .word   r15_to_r12
        .word   0
        .word   r13_to_r12
        .word   r14_to_r12
        .word   r15_to_r12

r13_to_r12:
        mov     r13, r12        /* get next word into r12       */
        b       check_a_bit

r14_to_r12:
        mov     r14, r12        /* get next word into r12       */
        b       check_a_bit

r15_to_r12:
        mov     r15, r12        /* get next word into r12       */
        b       check_a_bit

        /****************************************
         * We found a delayed task.  Now        *
         * decrement the tasks delay counter    *
         * and check to see if the counter is   *
         * zero.  If so, cleared the `delayed'  *
         * flag and set the `ready' flag.       *
         ***************************************/
delayed_task_found:
        mulo    r7, r6, r8      /* multiply 32 * [word]         */
        addo    r8, r5, r9      /* add bit offset               */
                                /* r9 contains priority         */
                                /* of delayed task              */
        ld      (r3)[r9*4], r8  /* r8 contains pointer          */
                                /* to TCB of delayed task       */
        ld      0x50(r8), r10   /* get current delay count      */
        subo    1, r10, r11     /* decrement delay count by one */
        st      r11, 0x50(r8)   /* update delay count           */

        cmpobl.f 0, r11, bump_bit_count
                                /* delay timed out?             */

        /****************************************
         * This task's delay count just timed   *
         * out.  Update its state from delayed  *
         * to running.                          *
         ***************************************/
        shro    3, r9, r8       /* byte to access in r8         */
        ldob    (r4)[r8*1], r10 /* get byte of delayed task     */
        and     7, r9, r11      /* get bit position for task    */
        clrbit  r11, r10, r9    /* clear delayed task bit       */
        stob    r9, (r4)[r8*1]  /* updated delay table          */
        ldob    _abTaskPrior[r8*1], r9
                                /* get byte in ready table      */
        setbit  r11, r9, r10    /* set task ready bit           */
        stob    r10, _abTaskPrior[r8*1]
                                /* update ready table           */
        b       bump_bit_count  /* check for more delayed tasks */

        /****************************************
         * All delayed tasks have been updated. *
         * Now determine if a context switch is *
         * necessary.                           *
         ***************************************/
delays_done:

        /* calculate the next highest priority task ready to run */
        ldconst _abTaskPrior, r3 /* base address of priority field */
        ldq     0x10(r3), r4     /* get first quad-word of array   */

        ldconst 7, r9            /* necessary constants for        */
        ldconst 32, r10          /* iterating through array        */

        scanbit r7, r8           /* look for set bit - MSB -> LSB  */
        be.t    rdy_task_found   /* ready task?                    */

        subo    1, r9, r9        /* decrement word counter         */
        scanbit r6, r8           /* look for set bit - MSB -> LSB  */
        be.t    rdy_task_found   /* ready task?                    */

        subo    1, r9, r9        /* decrement word counter         */
        scanbit r5, r8           /* look for set bit - MSB -> LSB  */
        be.t    rdy_task_found   /* ready task?                    */

        subo    1, r9, r9        /* decrement word counter         */
        scanbit r4, r8           /* look for set bit - MSB -> LSB  */
        be.t    rdy_task_found   /* ready task?                    */

        ldq     (r3), r4         /* load next quad-word            */

        subo    1, r9, r9        /* decrement word counter         */
        scanbit r7, r8           /* look for set bit - MSB -> LSB  */
        be.t    rdy_task_found   /* ready task?                    */

        subo    1, r9, r9        /* decrement word counter         */
        scanbit r6, r8           /* look for set bit - MSB -> LSB  */
        be.t    rdy_task_found   /* ready task?                    */

        subo    1, r9, r9        /* decrement word counter         */
        scanbit r5, r8           /* look for set bit - MSB -> LSB  */
        be.t    rdy_task_found   /* ready task?                    */

        subo    1, r9, r9        /* decrement word counter         */
        scanbit r4, r8           /* look for set bit - MSB -> LSB  */
        be.t    rdy_task_found   /* ready task?                    */

        ldconst 0, r5            /* we should never get here -     */
        b       check_for_switch /* should be at least idle task   */

        /****************************************
         * Found the highest priority task      *
         * ready to run.  Now calculate the     *
         * priority of that task.               *
         ***************************************/
rdy_task_found:
        mulo    r10, r9, r4      /* word count * 32                */
        addo    r4, r8, r5       /* plus bit position of set bit   */

        /****************************************
         * Is the highest priority task ready   *
         * to run already running?  If not,     *
         * switch.  If so, return.              *
         ***************************************/        
check_for_switch:
        ldob            _wHighestPriorRdy, r4
                                 /* get previous high prior task   */
        cmpobne.f       r4, r5, do_switch
                                 /* same as just calculated prior? */
        ret                      /* yes - continue!                */

        /****************************************
         * A context switch is warranted.  Set  *
         * up the stack to which the new task   *
         * will return.                         *
         *                                      *
         * This involves:                       *
         *   1) Get old PC and AC from inter-   *
         *      rupt record.  Store these       *
         *      values in pre-empted task's     *
         *      TCB.                            *
         *   2) Save the current state of TC    *
         *      (optional).                     *
         *   3) Save the previous frame         *
         *      pointer.                        *
         *   4) Save all global registers       *
         *      (notice global registers have   *
         *      not been used in this ISR).     *
         *   5) Update current task pointer.    *
         *   6) Flush the register conatined in *
         *      the on-chip register cache.     *
         *   7) Update the pfp to point to the  *
         *      new task's stack frame (the one *
         *      it will return to).             *
         *   8) Restore the state of the PC and *
         *      AC of the new task by writing   *
         *      the corresponding values        *
         *      contained in the task's TCB to  *
         *      the appropriate locations in    *
         *      interrupt record.               *
         *   9) Restore the state of the TC     *
         *      (optional).                     *
         *  10) Restore global registers.       *
         *  11) Execute a return instruction    *
         *      and voila...switch!             *
         ***************************************/

do_switch:
        st      r5, _wHighestPriorRdy
/* r5 : new wHighestPriorRdy */
                                 /* update wHighestPriorRdy        */
        ldconst _aptcbActiveTasks, r6
                                 /* base address of array of TCB * */
        ldconst TASK_READY, r3   /* constant indicating task state */
        ld      (r6)[r4*4], r7   /* get pointer to old task's TCB  */

        st      r3, 0x64(r7)     /* update the state of old task   */

        ld      -12(fp), r3      /* save old AC from int record    */
        st      r3, 0x48(r7)

        ld      -16(fp), r3      /* save old PC from int record    */
        st      r3, 0x4C(r7)

        /* save the current state of the TC */
        modtc   0, 0, r3
        st      r3, 0x44(r7)

        /* save the previous frame pointer */
        st      pfp, 0x40(r7)

        /* save the global registers  */
        /* don't worry about fp (g15) */
        /* it takes care of itself    */
        stq     g0,  (r7)
        stq     g4,  0x10(r7)
        stq     g8,  0x20(r7)
        stt     g12, 0x30(r7)

        /**************************************************
         * Now we are ready to restore the state of the   *
         * new task .                                     *
         *************************************************/

        ldconst TASK_RUNNING, r3 /* constant indicating task state */

        ld      (r6)[r5*4], r7   /* get pointer to new task        */
        st      r7, _ptcbCur     /* update pointer to current task */
        st      r3, 0x64(r7)     /* update the state of the task   */

        /* flush the local register frames and update pfp */
        flushreg
        ld      0x40(r7), pfp
        flushreg

        /****************************************
         * Again, second flushreg ensures pfp   *
         * is updated before any possibility of *
         * executing a return instruction.      *
         ***************************************/

        
        ld      0x48(r7), r3  /* restore the AC from current TCB   */
        st      r3, -12(fp)

        ld      0x4C(r7), r3  /* restore the PC from current TCB   */
        st      r3, -16(fp)

        ld      0x44(r7), r3  /* restore state of TC register      */
        ldconst 0x0FFE00FE, r4
        modtc   r4, r3, r5

        /* restore the state of the global registers */
        ldq     (r7), g0
        ldq     0x10(r7), g4
        ldq     0x20(r7), g8
        ldt     0x30(r7), g12

        /****************************************
         *      SWITCH TO NEW TASK !            *
         ***************************************/
        ret

_stack_exceeded:
.asciz  "\n\n***** Stack Overflow ***** : Task Priority %3d\n\n"

