/**************************************************
 *                                                *
 *  Module Name:  tomos.c                         *
 *                                                *
 *  Functions in this Module:                     *
 *    TOMOSInit()                                 *
 *    SchedTasks()                                *
 *    TaskDelete()                                *
 *    TaskDelay()                                 *
 *    TCBInit()                                   *
 *    TaskCreate()                                *
 *    StartTOMOS()                                *
 *    IdleTask()                                  *
 *                                                *
 *************************************************/

#include "ostypes.h"
#include "tomos.h"
#include "globals.h"
#include "utils.h"
#include "inline.h"

/**************************************************
 *                                                *
 * Function:  IdleTask()                          *
 *                                                *
 * Description:                                   *
 *   This is the background task TOMOS executes   *
 *   when no other task is ready to run.          *
 *                                                *
 * Prototype:                                     *
 *   void IdleTask(void *)                        *
 *                                                *
 * Parameters:                                    *
 *   None.                                        *
 *                                                *
 * Returns:                                       *
 *   Nothing                                      *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  01/13/95  Creation                       *
 *                                                *
 *************************************************/
void IdleTask(void *pv)
{
  int ii;

  /* turn on interrupts - begin task switching */
  EnableInts();

  for (ii = 0; ii < 1000000; ii++)
  {
    /* enter critical region */
    /* DisableInts(); */

    /* increment idle counter */
    wIdleCntr++;

    /* print a message so we know we're here */
    /* printf("tomos (IdleTask) : %05d\r", wIdleCntr); */

    /* end critical region */
    /* EnableInts(); */
  }

  /* code to exit from TOMOS - used for debug */
  printf("\n***** TOMOS Ending Now ***** %d\n", wIdleCntr);
  exit(1);
}


/**************************************************
 *                                                *
 * Function:  TOMOSInit()                         *
 *                                                *
 * Description:                                   *
 *  Initializes the data structures necessary for *
 *  TOMOS to start.  This function must be called *
 *  first, before any tasks are created or before *
 *  the call to StartTOMOS().                     *
 *                                                *
 * Prototype:                                     *
 *   void TOMOSInit(void)                         *
 *                                                *
 * Parameters:                                    *
 *   None.                                        *
 *                                                *
 * Returns:                                       *
 *   Nothing.                                     *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  12/31/94  Creation                       *
 *                                                *
 *************************************************/
void TOMOSInit(void)
{
  int ii;

  /* initialize OS state variables */
  wTicker          = 0;
  wHighestPriorRdy = IDLE_TASK_PRIORITY;
  wIdleCntr        = 0;

  /*******************************************************
   * Initialize task control block pointers.   ptcbCur   *
   * points to TCB of currently running task.  ptcbList  *
   * points to list of available TCBs                    *
   ******************************************************/
  ptcbCur  = (OSTCB *) 0;
  ptcbList = (OSTCB *) 0;

  /* we're not running...yet */
  fOSRunning = FALSE;

  /**********************************************
  * The next bit of code initializes various    *
  * arrays of control structures and pointers   *
  * in order to set up the task control manage- *
  * ment.                                       *
  **********************************************/
  
  /* initialize table of ready tasks to zero - no tasks ready */
  for (ii = 0; ii < MAX_PRIOR_BYTES; ii++)
    abTaskPrior[ii] = 0;
  
  /* initialize table of delayed tasks to zero - no tasks delayed */
  for (ii = 0; ii < MAX_PRIOR_BYTES; ii++)
    abTaskDelayed[ii] = 0;
 
  /* initialize array of pointers to TCBs of active tasks - no active tasks */
  for (ii = 0; ii < MAX_TASKS; ii++)
    aptcbActiveTasks[ii] = (OSTCB *) 0;

  /* create linked list of free TCBs */
  for (ii = 0; ii < MAX_TASKS; ii++)
    atcbTasksTbl[ii].NextTCB = &atcbTasksTbl[ii + 1];
  atcbTasksTbl[MAX_TASKS].NextTCB = (OSTCB *) 0;

  /* pointer to first free TCB - points to atcbTasksTbl[0] */
  ptcbFreeList = atcbTasksTbl;

  /* create the background (idle) task */
  TaskCreate(IdleTask, (void *) 0, IDLE_TASK_STK_SZ, IDLE_TASK_PRIORITY, TRUE);
}

/**************************************************
 *                                                *
 * Function:  TaskDelete()                        *
 *                                                *
 * Description:                                   *
 *   Deletes a task by deallocating the task's    *
 *   stack and task control block.  The ready     *
 *   tables are also updated.                     *
 *                                                *
 * Prototype:                                     *
 *   int TaskDelete(WORD iPrior)                  *
 *                                                *
 * Parameters:                                    *
 *   iPrior     The priority of the task to       *
 *              delete.                           *
 *                                                *
 * Returns:                                       *
 *   Nothing.                                     *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  01/01/95  Creation                       *
 *                                                *
 *************************************************/
void TaskDelete(int iPrior)
{
  OSTCB    *pTCB;
  BYTE     bTmp;

  /* make sure idle task not deleted */
  if (iPrior == IDLE_TASK_PRIORITY)
    return;

  /* begin critical section */
  DisableInts();

  /* check for existence of task */
  if ((pTCB = aptcbActiveTasks[iPrior]) == (OSTCB *) 0)
  {
    /* task at this priority doesn't exist */
    /* end critical section and return     */
    EnableInts();
    return;
  }

  /* remove the entry from the active tasks table */
  aptcbActiveTasks[iPrior] = (OSTCB *) 0;

  /* remove the TCB for this task from the chain of active tasks */
  if (pTCB->PrevTCB == (OSTCB *) 0)
  {
    pTCB->NextTCB->PrevTCB = (OSTCB *) 0;
    ptcbList = pTCB->NextTCB;
  }
  else
  {
    pTCB->PrevTCB->NextTCB = pTCB->NextTCB;
    pTCB->NextTCB->PrevTCB = pTCB->PrevTCB;
  }

  /* return TCB to free list */
  pTCB->NextTCB = ptcbFreeList;
  ptcbFreeList = pTCB;

  /* clear the bit in the priority field */
  /* corresponding to this priority      */
  bTmp = abTaskPrior[iPrior >> 3];
  bTmp &= (BYTE) ~(1 << (iPrior & 0x07));
  abTaskPrior[iPrior >> 3] = bTmp;

  /* release allocated memory for task's stack */
  free(pTCB->pvAllocStack);
    
  /* end critical section */
  EnableInts();
}

/**************************************************
 *                                                *
 * Function:  TaskDelay()                         *
 *                                                *
 * Description:                                   *
 *   Delay the current task for a specified       *
 *   number of clock ticks.                       *
 *                                                *
 * Prototype:                                     *
 *   void TaskDelay(int iDelay)                   *
 *                                                *
 * Parameters:                                    *
 *   iDelay     The number of clocks ticks for    *
 *              which to delay execution of the   *
 *              current task.  This value is      *
 *              decremented by one each time the  *
 *              timer tick ISR executes.          *
 *                                                *
 * Returns:                                       *
 *   Nothing.                                     *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  01/27/95  Creation                       *
 *                                                *
 *************************************************/
void TaskDelay(WORD wDelay)
{
  int  iPrior;
  BYTE bTmp;

  /* make sure there is a currently active task */
  if (ptcbCur != (OSTCB *) 0)
  {
    /* begin critical section */
    DisableInts();

    /* get the current task's priority */
    iPrior = ptcbCur->wPriority;

    /* clear corresponding bit in priority field for current task */
    bTmp = abTaskPrior[iPrior >> 3];
    bTmp &= (BYTE) ~(1 << (iPrior & 0x07));
    abTaskPrior[iPrior >> 3] = bTmp;

    /* set corresponding bit in delayed task field */
    bTmp = abTaskDelayed[iPrior >> 3];
    bTmp |= (BYTE) (1 << (iPrior & 0x07));
    abTaskDelayed[iPrior >> 3] = bTmp;

    /* update TCB for current task with the delay value */
    ptcbCur->wDelay = wDelay;
    
    /* end critical section */
    EnableInts();
  }
}

/**************************************************
 *                                                *
 * Function:  TCBInit()                           *
 *                                                *
 * Description:                                   *
 *   Initializes the task control block (TCB)     *
 *   structure for a newly created task.          *
 *                                                *
 * Prototype:                                     *
 *   int TCBInit(int iTaskPrior, int iStackSize)  *
 *                                                *
 * Parameters:                                    *
 *   iTaskPrior The priority of the task whose    *
 *              TCB will be initialized.          *
 *   iStackSize The amount of memory required for *
 *              the task's stack in bytes.  This  *
 *              function dynamically allocates    *
 *              the required memory.              *
 *                                                *
 *   pvData     Void pointer to any task specific *
 *              data.  The user must define this  *
 *              pointer and the data it points    *
 *              to.                               *
 *   fSupMode   Boolean flag indicates whether    *
 *              task being initialized will       *
 *              execute in user or supervisor     *
 *              mode.                             *
 *                                                *
 * Returns:                                       *
 *   Pointer to task control block, if            *
 *   sucessful.                                   *
 *                                                *
 *   NULL if memory cannot be allocated.          *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  12/31/94  Creation                       *
 *                                                *
 *************************************************/
static OSTCB *TCBInit(int iTaskPrior, int iStackSize, void *pvData, BOOL fSupMode)
{
  OSTCB *pTCB;
  WORD  wTmp;
  void  *pvCurFrm;

  /*******************************************************
   * Get a free task control block (TCB) from the free   *
   * store, and dynamically allocate the task's stack    *
   * from the heap.  If either of these two operations   *
   * fail, return (OSTCB *) 0.                           *
   ******************************************************/

  /* allocate a new TCB - if none available return unsuccessful */
  pTCB = ptcbFreeList;
  if (pTCB == (OSTCB *) 0)
    /* available TCBs exhausted - return unsuccessful */
    return (OSTCB*) 0;

  /* allocate memory for the task's stack */
  /* if heap exhausted return OSTCB * (0) */
  pTCB->pvAllocStack = malloc((size_t) iStackSize);
  if (!pTCB->pvAllocStack)
    /* heap exhausted - cannot allocate stack */
    /* return unsuccessful                    */
    return (OSTCB *) 0;

  /*******************************************************
   * TCB and stack allocated successfully.  Now update   *
   * the linked lists of free TCBs, to show one was just *
   * consumed.                                           *
   ******************************************************/

  /* update the available TCB list - one more TCB consumed */
  ptcbFreeList = pTCB->NextTCB;
  aptcbActiveTasks[iTaskPrior] = pTCB;

  /* update links in linked list of active tasks */
  pTCB->NextTCB   = ptcbList;
  pTCB->PrevTCB   = (OSTCB *) 0;

  if (ptcbList != (OSTCB *) 0)
    ptcbList->PrevTCB = pTCB;
  ptcbList = pTCB;

  /*******************************************************
   * Now initialize the TCB structure for this task.     *
   ******************************************************/
  
  /* initialize the TCB elements */
  pTCB->wPriority  = iTaskPrior;         /* task's priority */
  pTCB->tsStatus   = TASK_READY;         /* task ready to run */
  pTCB->wDelay     = 0;                  /* not delayed       */
  pTCB->wStackSize = (WORD) iStackSize;  /* task's stack size */

  /*******************************************************
   * The next step determines the address of the first   *
   * stack frame used by this task.  Recall that stack   *
   * frames -MUST- be located on quad-word (16-byte)     *
   * boundaries.  Each stack frame contains the local    *
   * register values, along with any local variable      *
   * values for that context.                            *
   ******************************************************/

  /* set up pointer to first stack frame of newly   */
  /* created task - make sure its quad-word aligned */
  wTmp = (WORD) pTCB->pvAllocStack;
  wTmp = ((wTmp + 0xF) & ~0xF);
  pvCurFrm = (void *) wTmp;

  /*******************************************************
   * Global variables must be saved across context       *
   * switches, becuase they are preserved across calls   *
   * and returns (even those to and from interrupts).    *
   * For the new task, initialize the TCB such that all  *
   * global registers will be initialized to zero        *
   * (except g0) when the task begins execution.  Global *
   * register g0 is the user defined pointer to task     *
   * specific data.                                      *
   ******************************************************/
  
  /* initial values for global registers */
  pTCB->GlobalRegs.g0  = (WORD) pvData;     /* user defined pointer to */
                                            /* task specific data      */
  pTCB->GlobalRegs.g1  = 0;
  pTCB->GlobalRegs.g2  = 0;
  pTCB->GlobalRegs.g3  = 0;
  pTCB->GlobalRegs.g4  = 0;
  pTCB->GlobalRegs.g5  = 0;
  pTCB->GlobalRegs.g6  = 0;
  pTCB->GlobalRegs.g7  = 0;
  pTCB->GlobalRegs.g8  = 0;
  pTCB->GlobalRegs.g9  = 0;
  pTCB->GlobalRegs.g10 = 0;
  pTCB->GlobalRegs.g11 = 0;
  pTCB->GlobalRegs.g12 = 0;
  pTCB->GlobalRegs.g13 = 0;
  pTCB->GlobalRegs.g14 = 0;
  pTCB->GlobalRegs.g15 = 0;

  /*******************************************************
   * Save the address of the first stack frame in the    *
   * task's allocated stack space.  When the task begins *
   * execution, it will start with this stack frame      *
   * because we will write the previous frame pointer    *
   * (pfp, or r0) with this value.  A return instruction *
   * will then be executed and the processor will return *
   * naturally to this stack frame (and context).        *
   *                                                     *
   * Recall that the previous frame pointer (pfp, or r0) *
   * points to the previously active stack frame, and    *
   * indicates the return type in the four least         *
   * significant bits (since all stack frames must be    *
   * quad-word aligned).                                 *
   *                                                     *
   * The return type field tells processor whether or    *
   * not it must get information from the interrupt      *
   * stack with which to restore the state upon return.  *
   * We will indicate a return type of interrupt-return  *
   * because the return from the timer tick interrupt    *
   * will be used to begin execution of this task.       *
   ******************************************************/
  wTmp = (WORD) pvCurFrm;
  wTmp &= ~0xF;                /* zero bits 0 thru 3                   */
  wTmp |= 0x07;                /* set bits 0 thru 2 - interrupt return */
  pTCB->wPrevFrmPtr = wTmp;

  /*******************************************************
   * Because the return type is interrupt-return, the    *
   * processor will restore the state of the arithmetic  *
   * controls (AC) and process controls (PC) registers   *
   * from the previous values stored on the interrupt    *
   * stack.                                              *
   *                                                     *
   * However, since this is a new task those values have *
   * not been written to the interrupt stack.  We will   *
   * have to do that when the time comes to begin        *
   * exeuction of this task.  For now, we must determine *
   * the values that -will- be written on the interrupt  *
   * stack, which will be used to restore the state of   *
   * of the AC and PC.                                   *
   *                                                     *
   * Also not we get the current value of the trace      *
   * controls (TC) register.  The TC register is not     *
   * strictly part of the state of the processor, but is *
   * optionally saved on a task-by-task basis.  This is  *
   * optional; it is not necessary to save the TC        *
   * register.  Also note that MON960 might get upset if *
   * you muck around with TC too much.                   *
   ******************************************************/

  /* store initial values of control registers */
  pTCB->TraceCtl = ReadTC();            /* get current value of TC    */
                                        /* can't change due to MON960 */
  pTCB->ArithCtl = ReadACInitImage();   /* read the AC initial image  */
                                        /* from the PRCB              */
  if (fSupMode)
    pTCB->ProcCtl  = INIT_PC;           /* reset value of PC - super mode task */
  else
    pTCB->ProcCtl  = INIT_PC - 2;       /* user mode task                      */

  /* success - return pointer to newly created TCB */
  return pTCB;
}

/**************************************************
 *                                                *
 * Function:  TaskCreate()                        *
 *                                                *
 * Description:                                   *
 *   Creates a task by allocating and init-       *
 *   ializing the task's stack.  Also updates the *
 *   task control block.                          *
 *                                                *
 * Prototype:                                     *
 *   int TaskCreate(void (*Task)(void *),         *
 *                  void *pData,                  *
 *                  int  iStackSize,              *
 *                  int  iTaskPrior,              *
 *                  BOOL fSupMode)                *
 *                                                *
 * Parameters:                                    *
 *   Task       Pointer to function taking void   *
 *              pointer argument and returning    *
 *              void.  This is the pointer to the *
 *              task function.                    *
 *   pData      Pointer to user defined data.     *
 *   iStackSize Size of the task's stack in       *
 *              bytes.                            *
 *   iTaskPrior The OS priority at which the task *
 *              will execute.                     *
 *   fSupMode   1: Create supervisor mode task.   *
 *              0: Create user mode task.         *
 *                                                *
 * Returns:                                       *
 *   SUCCESS,     if successful                   *
 *   FAILURE,     if task already exists at the   *
 *                the requested priority, or      *
 *                memory cannot be allocated.     *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  12/31/94  Creation                       *
 *                                                *
 *************************************************/
int TaskCreate(void (*Task)(void *), void *pvData, int iStackSize, int iTaskPrior, BOOL fSupMode)
{
  OSTCB *pTCB;
  BYTE  bPrior;
  int   ii;
  void  *pvCurFrm;
  WORD  wTmp;

  /* begin critical region */
  DisableInts();

  /* make sure task doesn't already exist at this priority */
  if (aptcbActiveTasks[iTaskPrior] != (OSTCB *) 0)
  {
    /* enable interrupts - return error */
    if (fOSRunning)
      EnableInts();
    return FAILURE;
  }

  /* intialize the TCB for this task */
  if ((pTCB = TCBInit(iTaskPrior, iStackSize, pvData, fSupMode)) == (OSTCB *) 0)
  {
    if (fOSRunning)
      EnableInts();
    return FAILURE;
  }

  /* get the address of the task's first stack frame */
  wTmp = pTCB->wPrevFrmPtr;
  wTmp &= ~0xF;
  pvCurFrm = (void *) wTmp;

  /*******************************************************
   * We now begin actually creating the stack frame in   *
   * external memory by initializing the locations which *
   * will hold the local registers to zero.  We will     *
   * then write the value of the stack pointer (sp, or   *
   * r1) and return instruction pointer (rip, or r2) in  *
   * their corresponding locations.                      *
   ******************************************************/
   
  /* initialize all local register locations to zero */
  for (ii = 0; ii < 16; ii++)
    *((WORD *) (pvCurFrm) + ii) = 0;

  /*******************************************************
   * Update stack pointer (r1) in new frame.  Allocate   *
   * space for global registers plus space for one word  *
   * value, which is the `void *' argument in the task   *
   * procedure.                                          *
   ******************************************************/
  ((LOCALREGS *) (pvCurFrm))->r1 = ((WORD) (pvCurFrm)) + 0x44;

  /*******************************************************
   * Now update the return instruction pointer (r2).     *
   * This should be the beginning address of the task    *
   * procedure.                                          *
   ******************************************************/
  ((LOCALREGS *) (pvCurFrm))->r2 = (WORD) Task;

  /* set corresponding bit in task priority field */
  bPrior = abTaskPrior[iTaskPrior >> 3];
  bPrior |= (BYTE) (1 << (iTaskPrior & 0x07));
  abTaskPrior[iTaskPrior >> 3] = bPrior;

  /* end critical section */
  if (fOSRunning)
    EnableInts();
  
  /* success */
  return SUCCESS;
}

/**************************************************
 *                                                *
 * Function:  StartTOMOS()                        *
 *                                                *
 * Description:                                   *
 *   Starts TOMOS multitasking kernel.  Calling   *
 *   this functions kicks off multitasking of all *
 *   currently created and ready tasks.           *
 *                                                *
 * Prototype:                                     *
 *   int StartTOMOS(void)                         *
 *                                                *
 * Parameters:                                    *
 *   None.                                        *
 *                                                *
 * Returns:                                       *
 *   Nothing                                      *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  01/11/95  Creation                       *
 *                                                *
 *************************************************/
void StartTOMOS()
{
  fOSRunning = TRUE;

  /* make idle task run first to enable interrupts */
  wHighestPriorRdy = IDLE_TASK_PRIORITY;

  /* kick off multitasking */
  StartHighestPriorRdy();
}

