/**************************************************
 *                                                *
 *  Module Name:  utils.c                         *
 *                                                *
 *  Functions in this Module:                     *
 *    GetPRCB()                                   *
 *    ReadIPND()                                  *
 *    ReadIMSK()                                  *
 *    WriteIPND()                                 *
 *    WriteIMSK()                                 *
 *    SetProcPriority()                           *
 *    ReadAC()                                    *
 *    WriteAC()                                   *
 *    ReadPC()                                    *
 *    WritePC()                                   *
 *    ReadTC()                                    *
 *    WriteTC()                                   *
 *    ReadACInitImage()                           *
 *    GetNextRdyTask()                            *
 *    FrameDump()                                 *
 *                                                *
 *************************************************/

#include "ostypes.h"
#include "arch960.h"
#include "globals.h"
#include "mmraddr.h"
#include "utils.h"
#include "inline.h"

/**************************************************
 *                                                *
 * Function:  GetPRCB()                           *
 *                                                *
 * Description:                                   *
 *   Uses a MON960 system call (for Cx) or reads  *
 *   the memory mapped register (for Jx or Hx) to *
 *   get the current pointer to the Processor     *
 *   Control Block (PRCB).                        *
 *                                                *
 * Prototype:  PRCB *GetPRCB(void)                *
 *                                                *
 * Parameters:                                    *
 *   None.                                        *
 *                                                *
 * Returns:                                       *
 *   The pointer to the current PRCB data         *
 *   structure.                                   *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  12/30/94  Creation                       *
 *                                                *
 *************************************************/
PRCB *GetPRCB(void)
{
  #if defined(__i960CA__) || defined(__i960CF__)

  /*******************************************************
   * Call the function provided by MON960 to get current *
   * PRCB pointer.  The user may override this function  *
   * as desired.                                         *
   ******************************************************/

  return (PRCB *) get_prcb();

  #elif defined(__i960JA__) || 
        defined(__i960JD__) ||
        defined(__i960JF__) ||
        defined(__i960HA__) || 
        defined(__i960HD__) ||
        defined(__i960HT__)

  /*******************************************************
   * Get the current PRCB pointer by reading the value   *
   * from the memory-mapped register.                    *
   ******************************************************/

  register PRCB *prcbptr;

  asm("lda PRCB_PTR, %0" : : "=&d" (prcbptr));
  asm("ld (%0), %0"      : : "=&d" (prcbptr));
  return prcbptr;
  
  #else
  #error Unsupported Architecture
  #endif
}

/**************************************************
 *                                                *
 * Function:  ReadIPND()                          *
 *                                                *
 * Description:                                   *
 *   Reads the interrupt pending register (IPND). *
 *                                                *
 * Prototype:  WORD ReadIPND(void)                *
 *                                                *
 * Parameters:                                    *
 *   None.                                        *
 *                                                *
 * Returns:                                       *
 *   The current value of the IPND register.      *
 *   This function should be called only when     *
 *   interrupts are disabled and the processor is *
 *   in supervisor mode.                          *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  12/30/94  Creation                       *
 *                                                *
 *************************************************/
WORD ReadIPND()
{
  register WORD wIPND;

  #if defined(__i960CA__) || defined(__i960CF__)

  /*******************************************************
   * Access the IPND register as a special function      *
   * register (sfr).                                     *
   ******************************************************/

  asm("mov sf0, %0" : "=&d" (wIPND));

  #elif defined(__i960JA__) ||
        defined(__i960JD__) ||
        defined(__i960JF__) ||
        defined(__i960JF__) ||
        defined(__i960HA__) || 
        defined(__i960HD__) ||
        defined(__i960HT__)

  /*******************************************************
   * Access the IPND register using its memory-mapped    *
   * address.                                            *
   ******************************************************/

  asm("ld IPND_ADDR, %0" : "=&d" (wIPND));
  
  #else
  #error Unsupported Architecture
  #endif

  /* return result */
  return wIPND;
}

/**************************************************
 *                                                *
 * Function:  ReadIMSK()                          *
 *                                                *
 * Description:                                   *
 *   Reads the interrupt mask register (IMSK).    *
 *                                                *
 * Prototype:  WORD ReadIMSK(void)                *
 *                                                *
 * Parameters:                                    *
 *   None.                                        *
 *                                                *
 * Returns:                                       *
 *   The current value of the IMSK register.      *
 *   This function should be called only when     *
 *   interrupts are disabled and the processor is *
 *   in supervisor mode.                          *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  12/30/94  Creation                       *
 *                                                *
 *************************************************/
WORD ReadIMSK()
{
  register WORD wIMSK;

  #if defined(__i960CA__) || defined(__i960CF__)

  /*******************************************************
   * Access the IMSK register as a special function      *
   * register (sfr).                                     *
   ******************************************************/

  asm("mov sf1, %0" : "=&d" (wIMSK));

  #elif defined(__i960JA__) ||
        defined(__i960JD__) ||
        defined(__i960JF__) ||
        defined(__i960HA__) || 
        defined(__i960HD__) ||
        defined(__i960HT__)

  /*******************************************************
   * Access the IMSK register using its memory-mapped    *
   * address.                                            *
   ******************************************************/

  asm("ld IMSK_ADDR, %0" : "=&d" (wIMSK));

  #else
  #error Unsupported Architecture
  #endif

  /* return result */
  return wIMSK;
}

/**************************************************
 *                                                *
 * Function:  WriteIPND()                         *
 *                                                *
 * Description:                                   *
 *   Writes the interrupt pending register        *
 *   (IPND).                                      *
 *                                                *
 * Prototype:  void WriteIPND(WORD wNewVal)       *
 *                                                *
 * Parameters:                                    *
 *   wNewVal   Contains the new value to write to *
 *             the IPND register.                 *
 *                                                *
 * Returns:                                       *
 *   Nothing.  This function should be called     *
 *   only when interrupts are disabled and the    *
 *   processor is in supervisor mode.             *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  12/30/94  Creation                       *
 *                                                *
 *************************************************/
void WriteIPND(WORD wNewVal)
{
  register WORD wIPND;

  #if defined(__i960CA__) || defined(__i960CF__)

  /*******************************************************
   * Access the IPND register as a special function      *
   * register (sfr).                                     *
   ******************************************************/

  asm("mov %0, sf0" : : "d" (wNewVal));

  #elif defined(__i960JA__) ||
        defined(__i960JD__) ||
        defined(__i960JF__) ||
        defined(__i960HA__) || 
        defined(__i960HD__) ||
        defined(__i960HT__)

  /*******************************************************
   * Access the IPND register using its memory-mapped    *
   * address.                                            *
   ******************************************************/

  asm("st %0, IPND_ADDR" : : "d" (wNewVal));

  #else
  #error Unsupported Architecture
  #endif
}

/**************************************************
 *                                                *
 * Function:  WriteIMSK()                         *
 *                                                *
 * Description:                                   *
 *   Writes the interrupt mask register           *
 *   (IMSK).                                      *
 *                                                *
 * Prototype:  void WriteIMSK(WORD wNewVal)       *
 *                                                *
 * Parameters:                                    *
 *   wNewVal   Contains the new value to write to *
 *             the IMSK register.                 *
 *                                                *
 * Returns:                                       *
 *   Nothing.  This function should be called     *
 *   only when interrupts are disabled and the    *
 *   processor is in supervisor mode.             *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  12/30/94  Creation                       *
 *                                                *
 *************************************************/
void WriteIMSK(WORD wNewVal)
{
  register WORD wIMSK;

  #if defined(__i960CA__) || defined(__i960CF__)

  /*******************************************************
   * Access the IMSK register as a special function      *
   * register (sfr).                                     *
   ******************************************************/

  asm("mov %0, sf1" : : "d" (wNewVal));

  #elif defined(__i960JA__) ||
        defined(__i960JD__) ||
        defined(__i960JF__) ||
        defined(__i960HA__) || 
        defined(__i960HD__) ||
        defined(__i960HT__)

  /*******************************************************
   * Access the IMSK register using its memory-mapped    *
   * address.                                            *
   ******************************************************/

  asm("st %0, IMSK_ADDR" : : "d" (wNewVal));

  #else
  #error Unsupported Architecture
  #endif
}

/**************************************************
 *                                                *
 * Function:  SetProcPriority()                   *
 *                                                *
 * Description:                                   *
 *   Sets the processor priority by updating the  *
 *   Process Controls (PC) register.              *
 *                                                *
 * Prototype:                                     *
 *   WORD SetProcPriority(WORD wNewPrior)         *
 *                                                *
 * Parameters:                                    *
 *   wNewPrior   Contains the new priority to     *
 *               the processor priority field of  *
 *               the PC register.  If called with *
 *               value of zero, PC is read only   *
 *               (no write).  This function       *
 *               should only be called when       *
 *               interrupts are disabled and the  *
 *               processor is in supervisor mode. *
 *                                                *
 * Returns:                                       *
 *   The previous processor priority.  Returns    *
 *   0xFFFFFFFF if unsuccessful.                  *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  12/30/94  Creation                       *
 *                                                *
 *************************************************/
WORD SetProcPriority(WORD wNewPrior)
{
  WORD wOldPrior, wMask;

  /* new priority must be between 0 and 31 */
  if (wNewPrior > 31)
    return 0xFFFFFFFF;

  /* mask allows update of only processor priority field */
  wMask = 0x001F0000;

  /* read the current value of the PC register */
  wOldPrior = 0;
  asm("modpc 0, 0, %0" : "=&d" (wOldPrior));

  /* only update priority if new priority not zero */
  if (wNewPrior)
  {
    /* now update PC register */
    wNewPrior <<= 16;
    asm("modpc 0, %0, %1" : : "d" (wMask), "d" (wNewPrior));
  }

  /* return the old processor priority */
  wOldPrior = (wOldPrior >> 16) & 0x1F;
  return wOldPrior;
}

/**************************************************
 *                                                *
 * Function:  ReadAC()                            *
 *                                                *
 * Description:                                   *
 *   Reads and returns the value of the           *
 *   Arithmetic Controls (AC) register.  Must be  *
 *   in supervisor mode to call this function.    *
 *                                                *
 * Prototype:                                     *
 *   WORD ReadAC(void)                            *
 *                                                *
 * Parameters:                                    *
 *   None.                                        *
 *                                                *
 * Returns:                                       *
 *   The value of the AC register.                *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  12/31/94  Creation                       *
 *                                                *
 *************************************************/
ARITHCTL ReadAC()
{
  ARITHCTL AC;

  /* mask is zero for read-only */
  AC.Value = 0;

  /* read the AC register */
  asm("modac 0, 0, %0" : "=&d" (AC.Value));

  /* return the value */
  return AC;
}

/**************************************************
 *                                                *
 * Function:  WriteAC()                           *
 *                                                *
 * Description:                                   *
 *   Writes the Atithmetic Controls register.     *
 *                                                *
 * Prototype:  ARITHCTL WriteAC(ARITHCTL AC)      *
 *                                                *
 * Parameters:                                    *
 *   AC        Contains the new value to write to *
 *             the AC register.                   *
 *                                                *
 * Returns:                                       *
 *   the value of the AC before modification      *
 *   occurred.                                    *
 *                                                *
 *   This function should be called               *
 *   only when interrupts are disabled and the    *
 *   processor is in supervisor mode.             *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  01/12/95  Creation                       *
 *                                                *
 *************************************************/
ARITHCTL WriteAC(ARITHCTL AC)
{
  WORD wMask;
  ARITHCTL OldAC;

  /* create the mask and write the AC register */
  /* the mask prevent writing to reserved bits */
  wMask = 0x00009107;
  asm("modpc %1, %2, %0" : "=&d" (OldAC.Value) : "dI" (wMask), "dI" (AC.Value));
  
  /* return the original value of the AC register */
  return OldAC;
}

/**************************************************
 *                                                *
 * Function:  ReadPC()                            *
 *                                                *
 * Description:                                   *
 *   Reads and returns the value of the           *
 *   Process Controls (PC) register.  Must be in  *
 *   supervisor mode to call this function.       *
 *                                                *
 * Prototype:                                     *
 *   PROCCTL ReadPC(void)                         *
 *                                                *
 * Parameters:                                    *
 *   None.                                        *
 *                                                *
 * Returns:                                       *
 *   The value of the PC register.                *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  12/31/94  Creation                       *
 *                                                *
 *************************************************/
PROCCTL ReadPC()
{
  PROCCTL PC;

  /* mask is zero for read-only */
  PC.Value = 0;

  /* read the PC register */
  asm("modpc 0, 0, %0" : "=&d" (PC.Value));

  /* return the value */
  return PC;
}

/**************************************************
 *                                                *
 * Function:  WritePC()                           *
 *                                                *
 * Description:                                   *
 *   Writes the Process Controls register.        *
 *                                                *
 * Prototype:  PROCCTL WritePC(PROCCTL PC)        *
 *                                                *
 * Parameters:                                    *
 *   PC        Contains the new value to write to *
 *             the PC register.                   *
 *                                                *
 * Returns:                                       *
 *   The value of the PC before modification      *
 *   occurred.                                    *
 *                                                *
 *   This function should be called               *
 *   only when interrupts are disabled and the    *
 *   processor is in supervisor mode.             *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  01/12/95  Creation                       *
 *                                                *
 *************************************************/
PROCCTL WritePC(register PROCCTL PC)
{
  WORD wMask;

  /* create the mask and write the PC register */
  /* mask prevents writing to reserved bits    */
  wMask = 0x001F2403;
  asm("modpc 0, %0, %1" : : "dI" (wMask), "dI" (PC.Value));
  
  /* return the original value of the AC register */
  return PC.Value;
}

/**************************************************
 *                                                *
 * Function:  ReadTC()                            *
 *                                                *
 * Description:                                   *
 *   Reads and returns the value of the           *
 *   Trace Controls (TC) register.  Must be in    *
 *   supervisor mode to call this function.       *
 *                                                *
 * Prototype:                                     *
 *   WORD ReadTC(void)                            *
 *                                                *
 * Parameters:                                    *
 *   None.                                        *
 *                                                *
 * Returns:                                       *
 *   The value of the TC register.                *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  12/31/94  Creation                       *
 *                                                *
 *************************************************/
TRCCTL ReadTC()
{
  TRCCTL TC;

  /* mask is zero for read-only */
  TC.Value = 0;

  /* read the AC register */
  asm("modtc 0, 0, %0" : "=&d" (TC.Value));

  /* return the value */
  return TC;
}

/**************************************************
 *                                                *
 * Function:  WriteTC()                           *
 *                                                *
 * Description:                                   *
 *   Writes the Trace Controls register.          *
 *                                                *
 * Prototype:  TRCCTL WriteTC(TRCCTL TC)          *
 *                                                *
 * Parameters:                                    *
 *   TC        Contains the new value to write to *
 *             the TC register.                   *
 *                                                *
 * Returns:                                       *
 *   The value of the TC before modification      *
 *   occurred.                                    *
 *                                                *
 *   This function should be called               *
 *   only when interrupts are disabled and the    *
 *   processor is in supervisor mode.             *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  01/12/95  Creation                       *
 *                                                *
 *************************************************/
TRCCTL WriteTC(TRCCTL TC)
{
  WORD wMask;
  TRCCTL OrgTC;

  /* create the mask and write the TC register */
  /* mask prevents writing to reserved bits    */
  wMask = 0x0FFE00FE;
  asm("modtc %1, %2, %0" : "=&d" (OrgTC.Value) : "dI" (wMask), "dI" (TC.Value));
  
  /* return the original value of the AC register */
  return OrgTC.Value;
}

/**************************************************
 *                                                *
 * Function:  ReadACInitImage()                   *
 *                                                *
 * Description:                                   *
 *   Reads and returns the value of the           *
 *   Arithmetic Controls (AC) register initial    *
 *   image found in the PRCB.                     *
 *                                                *
 * Prototype:                                     *
 *   ARITHCTL ReadACInitImage(void)               *
 *                                                *
 * Parameters:                                    *
 *   None.                                        *
 *                                                *
 * Returns:                                       *
 *   The value of the AC register initial image.  *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  12/31/94  Creation                       *
 *                                                *
 *************************************************/
ARITHCTL ReadACInitImage(void)
{
  PRCB *pPRCB;

  /* get the current PRCB pointer */
  pPRCB = GetPRCB();

  /* read and return the AC initial image */
  return pPRCB->AC;
}

/**************************************************
 *                                                *
 * Function:  GetNextRdyTask()                    *
 *                                                *
 *************************************************/
BYTE GetNextRdyTask()
{
  BYTE bTmp;
  WORD wTmp;

  /*******************************************************
   * The task priority field (abTaskPrior) is a bit      *
   * field which contains one bit for each task.  If a   *
   * task's bit is set, the task is ready to run.  If    *
   * not, the task either does not exist or is delayed.  *
   *                                                     *
   * The least significant bit of the least significant  *
   * word of this array corresponds to priority zero.    *
   * Similarly, the most significant bit of the most     *
   * significant word corresponds to the highest         *
   * priority task.                                      *
   *                                                     *
   * This code determines the highest priority task      *
   * which has its corresponding ready bit set, and      *
   * returns the priority of that task.                  *
   ******************************************************/
  
  for (wTmp = MAX_PRIOR_BYTES; wTmp > 0; wTmp--)
  {
    /* iterate through task priority field - highest byte first */
    bTmp = abTaskPrior[wTmp - 1];

    /* any bits set? at least one bit (for idle task) must be set */
    if (bTmp)
    {
      /* determine highest order bit set */
      if (bTmp & 0x80)
      {
        /* calculate priority */
        bTmp = (BYTE) ((8 * (wTmp - 1)) + 7);
        break;
      }

      if (bTmp & 0x40)
      {
        /* calculate priority */
        bTmp = (BYTE) ((8 * (wTmp - 1)) + 6);
        break;
      }
      
      if (bTmp & 0x20)
      {
        /* calculate priority */
        bTmp = (BYTE) ((8 * (wTmp - 1)) + 5);
        break;
      }
      
      if (bTmp & 0x10)
      {
        /* calculate priority */
        bTmp = (BYTE) ((8 * (wTmp - 1)) + 4);
        break;
      }
      
      if (bTmp & 0x08)
      {
        /* calculate priority */
        bTmp = (BYTE) ((8 * (wTmp - 1)) + 3);
        break;
      }
      
      if (bTmp & 0x04)
      {
        /* calculate priority */
        bTmp = (BYTE) ((8 * (wTmp - 1)) + 2);
        break;
      }
      
      if (bTmp & 0x02)
      {
        /* calculate priority */
        bTmp = (BYTE) ((8 * (wTmp - 1)) + 1);
        break;
      }
      
      if (bTmp & 0x01)
      {
        /* calculate priority */
        bTmp = (BYTE) (8 * (wTmp - 1));
        break;
      }
    }
  }

  /* return the priority of highest priority task ready to run */
  return bTmp;
}

/**************************************************
 *                                                *
 * Function:  FrameDump()                         *
 *                                                *
 * Description:                                   *
 *   Dumps the contents of memory where the frame *
 *   at the top of the stack has been written or  *
 *   created.  Dumps the value to be written to   *
 *   the pfp, sp, and rip.  Used for debug        *
 *   purposes.                                    *
 *                                                *
 * Prototype:                                     *
 *   void FrameDump(int, BOOL)                    *
 *                                                *
 * Parameters:                                    *
 *   iPrior    Priority of task of interest.      *
 *   fEnabInts Flag indicating whether or not to  *
 *             enable interrupts upon exit.  If   *
 *             TRUE, enable interrupts; otherwise *
 *             leave disabled.                    *
 * Returns:                                       *
 *   Nothing.                                     *
 *                                                *
 * Revision History:                              *
 *                                                *
 *   TJ  01/15/95  Creation                       *
 *                                                *
 *************************************************/
void FrameDump(int iPrior, BOOL fEnabInts)
{
  LOCALREGS *pLR;
  OSTCB     *pTCB;
  WORD      wTmp;

  /* Interupts are disabled so that we can (somewhat) */
  /* safely call printf.  This is a hack and will be  */
  /* improved in future revisions.                    */
  DisableInts();

  /* get pointer to TCB of given priority */
  pTCB = aptcbActiveTasks[iPrior];

  /* make sure frame has been assigned */
  if (!pTCB)
  {
    printf("Frame for Priority %03d Not assigned yet\n\n", iPrior);
    return;
  }

  /* get pointer to top of stack of given priority */
  wTmp = pTCB->wPrevFrmPtr;

  /* mask off return-type bits - store result */
  wTmp &= ~0xF;
  pLR = (LOCALREGS *) wTmp;

  /* dump the interesting components of the frame */
  printf("utils (FrameDump) : Frame at %p\n", pTCB->wPrevFrmPtr);
  printf("Priority %03d     : Frame at %p\n", iPrior, pLR);
  printf("pfp (r0) : %08X\n", pLR->r0);
  printf("sp  (r1) : %08X\n", pLR->r1);
  printf("rip (r2) : %08X\n\n", pLR->r2);

  /* turn interrupts back on if requested */
  if (fEnabInts)
    EnableInts();
}
